#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <unistd.h>

#include <time.h>
#include <errno.h>

#include "const.h"
#include "types.h"
#include "engine.h"
#include "util.h"
#include "display.h"

const int n_gems = 7;
const long delay = 200;

void random_fill(
    board_t board,
    int size,
    int zeros_only)
{
    for (int i = 0; i < size; i++)
        for (int j = 0; j < size; j++)
            if (!zeros_only || !board[i][j])
                board[i][j] = rand() % n_gems + 1;
}

int random_fill_top_zeros(
    board_t board,
    int size)
{
    int zeros_exist = 0;
    for (int j = 0; j < size; j++)
        if (!board[0][j])
        {
            board[0][j] = rand() % n_gems + 1;
            zeros_exist = 1;
        }
    return zeros_exist;
}

void display(
    board_t board,
    int size)
{
    for (int i = 0; i < size; i++)
    {
        for (int j = 0; j < size; j++)
            printf("%2d ", board[i][j]);
        printf("\n");
    }
}

void init_board(
    board_t board,
    int size,
    int min_gems)
{
    srand((unsigned int)time(NULL));

    random_fill(board, size, 0);

    int hfound, vfound;

    while (flip(board, size) > 0)
    {
        cleanup(board, size);
        while (collapse(board, size) > 0)
            ;
        random_fill(board, size, 1);
    }
}

int update_board(
    state_t *state)
{
    int score_increase = flip(state->board, state->size);
    if (score_increase > 0)
    {
        cleanup(state->board, state->size);
        append_record(state->board, state->size);

        // Valid move!
        state->score += score_increase;
        return 1;
    }

    return 0;
}
void init_gem_colors()
{
    init_pair(1, COLOR_BLACK, COLOR_GREEN);
    init_pair(2, COLOR_BLACK, COLOR_BLUE);
    init_pair(3, COLOR_BLACK, COLOR_WHITE);
    init_pair(4, COLOR_BLACK, COLOR_MAGENTA);
    init_pair(5, COLOR_BLACK, COLOR_RED);
    init_pair(6, COLOR_BLACK, COLOR_YELLOW);
    init_pair(7, COLOR_BLACK, COLOR_CYAN);
}

int init_state(
    state_t *state,
    int argc,
    char **argv)
{
    state->size = 10;
    state->highlight_r = state->highlight_c = 0;
    state->choice_r = state->choice_c = -1;
    state->quit = 0;
    state->game_over = 0;
    state->active_selection = 0;
    state->swap = SWAP_NULL;
    state->collapse_animation = 0;
    state->new_gems_animation = 0;
    state->autopilot = 0;
    state->animation_mode = 0;
    state->debug = 0;
    state->score = 0;
    state->delay = 200;

    int c;
    while ((c = getopt(argc, argv, "Ds:hr:")) != -1)
    {
        switch (c)
        {
        case 'L':
            state->delay = atoi(optarg);
            break;
        case 'D':
            state->debug = 1;
            break;
        case 's':
            state->size = atoi(optarg);
            if (state->size == 0)
            {
                state->size = 10;
                printf("Cannot parse size\n");
            }
            printf("Size = %d\n", state->size);
            break;
        case 'r':
            if (!init_recording(optarg))
            {
                printf("Cannot open recording file %s\n", optarg);
                return 1;
            }
            break;
        case 'h':
        case '?':
            break;
        }
    }
    state->board = (board_t)malloc(state->size * sizeof(int *));
    for (int i = 0; i < state->size; i++)
    {
        state->board[i] = (int *)malloc(state->size * sizeof(int));
    }

    return 0;
}

void debug(state_t *state, const char *message)
{
    move(4, 0);
    clrtoeol();
    mvprintw(4, 0, message);

    if (state->debug)
    {
        refresh();
        wgetch(state->menu_win);
    }
    else
    {
        msleep(state->delay);
    }
}

void do_animate(
    state_t *state)
{
    if (state->swap && state->active_selection)
    {
        debug(state, "Trying swap");
        if (!update_board(state))
        {
            // Swap back
            debug(state, "Invalid swap");
            do_swap(state->board, state->size, state->swap, state->choice_r, state->choice_c);
            state->animation_mode = 0;
        }
        else
        {
            state->collapse_animation = 1;
        }
        state->swap = SWAP_NULL;
        state->active_selection = 0;
    }
    else if (state->collapse_animation)
    {
        debug(state, "Collapse");
        state->collapse_animation = collapse(state->board, state->size) > 0;
        if (!state->collapse_animation)
        {
            // Maybe a new combination has appeared
            if (update_board(state))
            {
                // Start again!
                state->collapse_animation = 1;
            }
            else
            {
                // collapse_animation is 0
                if (random_fill_top_zeros(state->board, state->size))
                {
                    state->new_gems_animation = 1;
                }
            }
            msleep(delay);
        }
    }
    else if (state->new_gems_animation)
    {
        debug(state, "New gems");
        state->new_gems_animation = collapse(state->board, state->size) > 0;
        if (!state->new_gems_animation)
        {
            // Maybe a new combination has appeared
            if (update_board(state))
            {
                state->collapse_animation = 1;
            }
            msleep(delay);
        }
        else
        {
            random_fill_top_zeros(state->board, state->size);
        }
    }
    else
    {
        debug(state, "Animation done");
        state->animation_mode = 0;
    }
}

void toggle_autopilot(state_t*state)
{
    if (state->autopilot)
    {
        // Blocking getch
        wtimeout(state->menu_win, -1);
        state->autopilot = 0;
        debug(state, "Autopilot off");
    }
    else 
    {
        // Non-blocking getch
        wtimeout(state->menu_win, 100);
        state->autopilot = 1;
        debug(state, "Autopilot on");
    }
}

void process_keys(
    state_t *state)
{
    int swap, row, col;
    int c = wgetch(state->menu_win);
    int exists;

    if (c == 'A')
    {
        toggle_autopilot(state);
    }

    if (state->autopilot || c == 'a')
    {
        exists = suggest_move(state->board, state->size, &row, &col, &swap);
        if (exists > 0)
        {
            do_swap(state->board, state->size, swap, row, col);
            state->animation_mode = 1;
            state->collapse_animation = 1;
        }
        else if (!exists)
        {
            state->game_over = 1;
            state->autopilot = 0;
        }
        else 
        {
            debug(state, "Next move not implemented!");
            state->autopilot = 0;
        }
        return;
    }

    switch (c)
    {
    case KEY_UP:
        if (state->active_selection)
        {
            state->swap = do_swap(state->board, state->size, SWAP_UP, state->choice_r, state->choice_c);
            state->animation_mode = state->swap != SWAP_NULL;
        }
        else
        {
            if (!state->highlight_r)
                state->highlight_r = state->size - 1;
            else
                --state->highlight_r;
        }
        break;

    case KEY_DOWN:
        if (state->active_selection)
        {
            state->swap = do_swap(state->board, state->size, SWAP_DOWN, state->choice_r, state->choice_c);
            state->animation_mode = state->swap != SWAP_NULL;
        }
        else
        {
            if (state->highlight_r == state->size - 1)
                state->highlight_r = 0;
            else
                ++state->highlight_r;
        }
        break;

    case KEY_LEFT:
        if (state->active_selection)
        {
            state->swap = do_swap(state->board, state->size, SWAP_LEFT, state->choice_r, state->choice_c);
            state->animation_mode = state->swap != SWAP_NULL;
        }
        else
        {
            if (!state->highlight_c)
                state->highlight_c = state->size - 1;
            else
                --state->highlight_c;
        }
        break;

    case KEY_RIGHT:
        if (state->active_selection)
        {
            state->swap = do_swap(state->board, state->size, SWAP_RIGHT, state->choice_r, state->choice_c);
            state->animation_mode = state->swap != SWAP_NULL;
        }
        else
        {
            if (state->highlight_c == state->size - 1)
                state->highlight_c = 0;
            else
                ++state->highlight_c;
        }
        break;

    case 'E':
    case 'e':
        c = wgetch(state->menu_win);
        int digit = c - '0';
        if (digit >= 1 && digit <= n_gems)
        {
            state->board[state->highlight_r][state->highlight_c] = digit;
            state->animation_mode = 1;
            state->collapse_animation = 1;
        }
        break;

    case 'D':
    case 'd':
        state->debug = 1 - state->debug;
        break;

    case '?':
        exists = suggest_move(state->board, state->size, &row, &col, &swap);
        if (exists > 0)
        {
            state->highlight_r = row;
            state->highlight_c = col;
        }
        else if (!exists)
        {
            state->game_over = 1;
        }
        else
        {
            debug(state, "Next move not implemented!");
        }
        break;

    case 10:
        if (!state->active_selection)
        {
            state->choice_r = state->highlight_r;
            state->choice_c = state->highlight_c;
            state->active_selection = 1;
        }
        else
        {
            // Toggle active selection
            state->active_selection = 0;
        }
        break;

    case 27:
    case 'q':
    case 'Q':
        state->quit = 1;
        break;
    }
}

int main(int argc, char **argv)
{
    int c;
    state_t state;

    if (init_state(&state, argc, argv))
    {
        return 1;
    }
    init_board(state.board, state.size, MIN_GEMS);
    printf("Board initialized\n");

    initscr();
    clear();
    noecho();
    cbreak(); /* Line buffering disabled. pass on everything */
    int startx = 5;
    int starty = 5;

    start_color();
    init_gem_colors();

    state.menu_win = newwin(state.size + 4, state.size * 3 + 4, starty, startx);
    keypad(state.menu_win, TRUE);
    mvprintw(0, 0, "Use arrow keys to navigate, Press enter to select a choice.");
    mvprintw(1, 0, "Once you make a choice press again one of the arrow keys.");
    mvprintw(2, 0, "Score: %6d", state.score);
    refresh();
    print_board(&state);
    while (!state.quit)
    {
        if (state.animation_mode)
        {
            do_animate(&state);
        }
        else
        {
            process_keys(&state);
        }
        mvprintw(2, 0, "Score: %6d %s", state.score, state.game_over ? "*** GAME OVER ***" : "");
        move(3, 0);
        clrtoeol();
        if (state.active_selection)
        {
            mvprintw(3, 0, "You have selected %d, %d", state.choice_r, state.choice_c);
        }
        else
        {
            mvprintw(3, 0, "No active selection");
        }
        print_board(&state);
        refresh();
    }
    endwin();
    terminate_recording();

    for (int i = 0; i < state.size; i++)
    {
        free(state.board[i]);
    }
    free(state.board);

    return 0;
}
